# 5 avatars : 92 fps
# 9 avatars : 60 fps
# 20 avatars : 30 fps
# 40 avatars : 16 fps
from pandac.PandaModules import *
import direct.directbase.DirectStart
from direct.gui.OnscreenText import OnscreenText
from direct.showbase.DirectObject import DirectObject
from direct.interval.IntervalGlobal import *
from direct.task import Task
from random import *
import sys

class myAvatar:
  def __init__(self,tex,scale,groundGeom):
      self.obj=loader.loadModelCopy('misc/Dirlight')
      self.obj.reparentTo(render)
      self.obj.find('**/Box03').setP(180)
      self.obj.flattenStrong()
      self.obj.setPosHprScale(uniform(-.1,.1),uniform(-.1,.1),-.15,  random()*360,0,0,  .05+.25*scale,.05+.25*scale,.04)
      self.obj.setColor(uniform(.3,1),uniform(.3,1),uniform(.3,1))
      self.tex=tex
      self.scale=1./scale
      self.groundGeom=groundGeom
      # basically this is what we need : (scale * .5) * 1.42, or scale * .71
        # (scale * .5) is the drop shadow's radius
        # 1.42 is sqrt(2), which is the distance from center to corner of the square texture
      # but for safety precaution, let's expand it a little :
      # the .05 constant is to fill in the gap between the geom's edge and the bounds,
      # adjust it based on your ground geom size
      self.dropShadowGeomRange=groundGeomXBounds*scale*.71+.03
      self.dropShadowGeom=dropShadowGeomRoot.attachNewNode('')

  def updateDropShadow(self,dt,debug=0):
      if SEE_DS_GEOM and not debug:
         return
      pos=self.obj.getPos()
      self.obj.setH(self.obj,uniform(-5,5))
      self.obj.setY(self.obj,uniform(dt*.5,dt))
      self.avatarPos=self.obj.getPos(self.groundGeom)
      if self.avatarPos.lengthSquared()>MOVEMENT_RANGE:
         self.obj.setH(self.obj,randint(-1,1)*75)
         self.obj.setPos(pos)
         self.avatarPos=pos
      
      self.dropShadowGeom.removeNode()
      self.dropShadowGeom=dropShadowGeomRoot.attachNewNode('')

      self.recurseFindGeom(self.groundGeom)

      DStexStage=TextureStage('')
      self.dropShadowGeom.setTexGen(DStexStage,RenderAttrib.MWorldPosition,1)
      self.dropShadowGeom.setTexture(DStexStage,self.tex,1)
      self.dropShadowGeom.setTexScale(DStexStage, self.scale)
      self.dropShadowGeom.setTexOffset(DStexStage,
         .5-self.avatarPos[0]*self.scale,
         .5-self.avatarPos[1]*self.scale)
      if not (SEE_DS_GEOM and debug):
         self.dropShadowGeom.setTransparency(TransparencyAttrib.MAlpha)

  def recurseFindGeom(self,groundGeom):
      nc=groundGeom.getNumChildren()
      if nc==0:
         if (self.avatarPos-groundGeom.getBounds().getCenter()).length() < groundGeom.getBounds().getRadius()+self.dropShadowGeomRange:
            np=groundGeom.instanceTo(self.dropShadowGeom)
         return

      for i in range(nc):
          self.recurseFindGeom(groundGeom.getChild(i))

class World(DirectObject):
  def __init__(self):
      self.renderMode=(RenderModeAttrib.MFilled,RenderModeAttrib.MWireframe,RenderModeAttrib.MPoint)
      self.renderModeIdx=0

      self.accept('escape',sys.exit)
      self.accept('r',self.changeRenderMode)
      self.accept('a',render.analyze)
      self.accept('space',self.toggleUpdatingAvatars)
      self.accept('enter',self.toggleDropShadowGeom)
      self.accept('arrow_left',self.prevAvatar)
      self.accept('arrow_right',self.nextAvatar)

      __builtins__.SEE_DS_GEOM=0
      __builtins__.AVATAR_IDX=0

      OnscreenText( text='[ R ] :: cycle render mode', scale=.05, pos=(0,-.84), fg=(1,1,1,1), shadow=(0,0,0,1) )
      OnscreenText( text='[ SPACE ] :: toggle movement', scale=.05, pos=(0,-.9), fg=(1,1,1,1), shadow=(0,0,0,1) )
      OnscreenText( text='[ ENTER ] :: toggle drop shadow geom', scale=.05, pos=(0,-.96), fg=(1,1,1,1), shadow=(0,0,0,1) )

      base.trackball.node().setForwardScale(.01)
      base.setBackgroundColor(0,0,0,1)
      base.camLens.setNearFar(.1,100000)
      camera.setPos(-.5,-1.5,.8)
      camera.lookAt(render)
      mat4=Mat4(camera.getMat())
      mat4.invertInPlace()
      base.mouseInterfaceNode.setMat(mat4)

      # create ground
      self.ripple=loader.loadModel('ripple')
      self.ripple.reparentTo(render)
      rippleGeom=self.ripple.find('**/+GeomNode') # the geom's CS is Y-up
      rippleGeom.setP(-90) # force it to be Z-up
      self.ripple.flattenStrong() # and save changes to vertices
      __builtins__.MOVEMENT_RANGE=.18
      # re-texturize the ground
      tex0=loader.loadTexture('maps/envir-ground.jpg')
      tex0.setMagfilter(Texture.FTLinearMipmapLinear)
      tex0.setMinfilter(Texture.FTLinearMipmapLinear)
      self.ripple.setTexture(tex0,1)

      # loads the special made (octree) ground, to which the drop shadows get applied,
      # make sure this model's color is white (1,1,1,1)
      self.rippleDropShadowGeom=loader.loadModel('ripple-split')
      self.rippleDropShadowGeom.reparentTo(render)
      self.rippleDropShadowGeom.setP(-90)
      self.rippleDropShadowGeom.flattenLight()
      self.rippleDropShadowGeom.hide()
      b0,b1=self.rippleDropShadowGeom.getTightBounds()
      __builtins__.groundGeomXBounds=(b1-b0)[0]
#       geomList=self.rippleDropShadowGeom.findAllMatches('**/+GeomNode').asList()
#       for g in geomList:
#           g.showBounds()

      __builtins__.dropShadowGeomRoot=render.attachNewNode('')
      dropShadowGeomRoot.setZ(.0002)
      dropShadowGeomRoot.setLightOff()
      dropShadowGeomRoot.setAttrib(DepthOffsetAttrib.make())
      dropShadowGeomRoot.setDepthWrite(0)

      # let's clip the unwanted parts below XY plane
      XYPlane = Plane( Vec3( 0, 0, 1 ), Point3( 0, 0, -.025 ) )
      XYPlaneNode = PlaneNode( 'positiveZ' )
      XYPlaneNode.setPlane( XYPlane )
      XYPlaneNP = render.attachNewNode( XYPlaneNode )
      render.setClipPlane(XYPlaneNP)

      self.createAvatars()
      taskMgr.add(self.updateAllAvatars,'avatarUpdateTask')
      self.updatingAvatars=1
      self.setupLights()

  def createAvatars(self):
      # the overlaying texture (just 1 for all now)
      tex1=loader.loadTexture('drop2.png')
      # clamp the texture borders
      tex1.setWrapU(Texture.WMClamp)
      tex1.setWrapV(Texture.WMClamp)

      # myAvatar class instances collection
      self.myAvatars=[]

      scale=.3
      self.myAvatars.append(
         myAvatar(tex1,scale,self.rippleDropShadowGeom)
         )

      scale=.2
      for i in range(4):
          self.myAvatars.append(
             myAvatar(tex1,scale,self.rippleDropShadowGeom)
             )

      scale=.12
      for i in range(4):
          self.myAvatars.append(
             myAvatar(tex1,scale,self.rippleDropShadowGeom)
             )

      for i in range(11):
          self.myAvatars.append(
             myAvatar(tex1,uniform(.05,.12),self.rippleDropShadowGeom)
             )
      print len(self.myAvatars)

  def updateAllAvatars(self,t):
      if self.updatingAvatars:
         dt=globalClock.getDt()*2
         idx=0
         for a in self.myAvatars:
             a.updateDropShadow( dt,debug=(AVATAR_IDX==idx) )
             idx+=1
      return Task.cont

  def toggleUpdatingAvatars(self):
      self.updatingAvatars=not self.updatingAvatars

  def toggleDropShadowGeom(self):
      __builtins__.SEE_DS_GEOM = not __builtins__.SEE_DS_GEOM
      if SEE_DS_GEOM:
         self.cycleAvatarText = OnscreenText( text='use Left/Right arrow keys to cycle through avatars',
                              scale=.05, pos=(0,-.7), fg=(1,1,1,1), shadow=(0,0,0,1) )
         __builtins__.AVATAR_IDX=0
         for a in self.myAvatars:
             a.dropShadowGeom.removeNode()
             a.obj.hide()
      else:
         self.cycleAvatarText.destroy()
         for a in self.myAvatars:
             a.obj.show()

  def nextAvatar(self):
      if not SEE_DS_GEOM:
         return
      self.myAvatars[AVATAR_IDX].obj.hide()
      self.myAvatars[AVATAR_IDX].dropShadowGeom.removeNode()
      __builtins__.AVATAR_IDX+=1
      __builtins__.AVATAR_IDX%=len(self.myAvatars)
  
  def prevAvatar(self):
      if not SEE_DS_GEOM:
         return
      self.myAvatars[AVATAR_IDX].obj.hide()
      self.myAvatars[AVATAR_IDX].dropShadowGeom.removeNode()
      __builtins__.AVATAR_IDX-=1
      __builtins__.AVATAR_IDX%=len(self.myAvatars)

  def changeRenderMode(self):
      self.renderModeIdx+=1
      self.renderModeIdx%=len(self.renderMode)
      render.setRenderMode(self.renderMode[self.renderModeIdx],1)

  def setupLights(self):
      ambientLight = AmbientLight( 'ambientLight' )
      ambientLight.setColor( Vec4(0.2, 0.2, 0.2, 1) )
      ambientLightNP=render.attachNewNode( ambientLight )

      directionalLight = DirectionalLight( 'directionalLight1' )
      directionalLight.setDirection( Vec3( -1, 0, -1 ) )
      directionalLight.setColor( Vec4( .8, .8, .8, 1 ) )
      directionalLightNP=render.attachNewNode( directionalLight )

      render.setLight(ambientLightNP)
      render.setLight(directionalLightNP)

World()
run()
